import logging
import time

import slixmpp

from .common import get_config


class ConnectionLost(Exception):
    pass


class BaseBot(slixmpp.ClientXMPP):
    def __init__(self, config_path):
        self.config = get_config(config_path)
        super().__init__(self.config['user'], self.config['pass'])
        logging.basicConfig(level=self.config.get(
            'LOG', logging.INFO), format='%(levelname)-8s %(message)s')
        self.log = logging.getLogger()

    def run(self, loop=True):
        while True:
            self.connect()
            self.log.info("Bot started.")
            self.process()
            if not loop:
                return
            time.sleep(5)

    def connection_lost(self, *args, **kargv):
        super().connection_lost(*args, **kargv)
        self.disconnect()



import time
from os.path import expanduser

import slixmpp
from bs4 import BeautifulSoup, Tag

from .basebot import BaseBot
from .common import get_config, to_xep0393


class SendMsgBot(BaseBot):
    def __init__(self, config_path):
        super().__init__(config_path)
        self.messages = []
        self.add_event_handler("session_start", self.start)

    def start(self, event):
        self.send_presence()
        self.get_roster()
        while self.messages:
            to, msg = self.messages.pop(0)
            self.send_message(mto=to,
                              mbody=msg,
                              mtype='chat')
            time.sleep(0.1)
        self.disconnect()

    def run(self):
        if not self.messages:
            return
        with timeout(seconds=10):
            try:
                super().run(loop=False)
            except TimeoutError:
                pass


class XmppMsg:
    def __init__(self, config=expanduser("~/.xmpp.yml"), to=None):
        self._config = None
        self._to = None
        self.config = config
        if to is not None:
            self.to = to

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, config):
        self.bot = SendMsgBot(config)
        self.to = self._config.get("to", self.to)

    @property
    def to(self):
        return self._to

    @to.setter
    def to(self, to):
        if to is None:
            self._to = None
        elif isinstance(to, str):
            self._to = to.strip().split()
        elif isinstance(to, (set, list, tuple)):
            self._to = to
        else:
            raise Exception("to must be a str, set, list or tuple")

    @property
    def msg(self):
        return self.bot.messages

    @msg.setter
    def msg(self, value):
        if value is None:
            self.bot.messages = []
            return
        if isinstance(value, (Tag, BeautifulSoup, str)):
            value = msg_to_xep0393(value)
        if isinstance(value, str):
            for to in self.to:
                self.bot.messages.append((to, value))
        elif isinstance(value, tuple):
            msg = msg_to_xep0393(value[-1])
            for tos in value[:-1]:
                for to in tos.strip().split():
                    self.bot.messages.append((to, msg))

    def send(self):
        self.bot.run()

    def reload(self, config=None):
        self.bot = SendMsgBot(config or self.bot.config)


def msg_to_xep0393(msg):
    if isinstance(msg, (Tag, BeautifulSoup)):
        return to_xep0393(msg)
    if isinstance(msg, str):
        _msg = msg.strip().lower()
        for tag in ("html", "p", "div", "body", "table"):
            if _msg.startswith("<"+tag+">") and _msg.endswith("</"+tag+">"):
                return to_xep0393(msg)
    return msg



import os
from setuptools import find_packages, setup

abspath = os.path.abspath(__file__)
dname = os.path.dirname(abspath)
os.chdir(dname)

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()
long_description = long_description.strip()

setup(
    name='slixmppbot',
    packages=find_packages(),
    version='1.0.3',
    long_description=long_description,
    long_description_content_type="text/markdown",
    description='A framework for writing Jabber/XMPP bots',
    author='s-nt-s',
    author_email='santos82h@gmail.com',
    url='https://github.com/s-nt-s/XmppBot',
    keywords=['xmpp', 'bot'],
    license='GPLv3',
    classifiers=[
        "Programming Language :: Python :: 3",
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5'
)



# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

"""
A framework for writing Jabber/XMPP bots.
The XmppBot framework allows you to easily write bots
that use the XMPP protocol. You can create commands by
decorating functions in your subclass or customize the
bot's operation completely.
"""

import inspect
import os
import re
import sys

import slixmpp

from .basebot import BaseBot

sp = re.compile(r"\s+", re.MULTILINE | re.UNICODE)
url_img = re.compile(r"(https?://\S+\.(gif|png|jpe?g)\S*)", re.IGNORECASE)

if sys.version_info < (3, 0):
    reload(sys)
    sys.setdefaultencoding('utf8')

creator_order = 0


def botcmd(*args, **kwargs):
    """Decorator for bot command functions"""
    global creator_order

    def decorate(func, creator_order, name=None, names=None, delay=False, fromuser=None, regex=None, rg_mode="match"):
        setattr(func, '_command', True)
        setattr(func, '_command_names', names or [name or func.__name__])
        setattr(func, '_command_delay', delay)
        setattr(func, '_command_fromuser', fromuser)
        setattr(func, '_command_regex', regex)
        setattr(func, '_command_rg_mode', rg_mode)
        setattr(func, '_command_order', creator_order)
        return func

    creator_order += 1

    if len(args):
        return decorate(args[0], creator_order, **kwargs)
    else:
        return lambda func: decorate(func, creator_order, **kwargs)


class XmppBot(BaseBot):
    MSG_ERROR_OCCURRED = "ERROR!!"

    def __init__(self, config_path):
        super().__init__(config_path)
        self.use_ipv6 = self.config.get("use_ipv6", True)
        self.delay = False
        self.commands = []
        rg_commands = []
        plugins = set(self.config.get("plugins", "").split())

        commands = inspect.getmembers(self, inspect.ismethod)
        commands = filter(lambda x: getattr(x[1], '_command', False), commands)
        commands = sorted(
            commands, key=lambda x: getattr(x[1], '_command_order'))

        for name, value in commands:
            names = getattr(value, '_command_names')
            order = getattr(value, '_command_order', 0)
            self.log.info('Registered %dº command: %s' %
                          (order, " ".join(names)))
            self.commands.append(value)
            self.delay = self.delay or getattr(value, '_command_delay', False)

        self.nick = self.config['user'].split("@")[0]

        self.auto_reconnect = True
        if self.config.get("auto", False):
            self.auto_authorize = True
            self.auto_subscribe = True

        plugins.add('xep_0030')  # Service Discovery
        plugins.add('xep_0004')  # Data Forms
        plugins.add('xep_0060')  # PubSub
        plugins.add('xep_0199')  # XMPP Ping
        if self.delay:
            plugins.add('xep_0203')  # XMPP Delayed messages
        if self.config.get('vcard', None):
            plugins.add('xep_0054')
        if self.config.get('avatar', None):
            if os.path.isfile(self.config['avatar']):
                plugins.add('xep_0084')
                plugins.add('xep_0153')
            else:
                del self.config['avatar']

        if self.config.get('rooms', None):
            plugins.add('xep_0045')  # Multi-User Chat

        if self.config.get('img_to_oob', False):
            plugins.add('xep_0066')  # OOB

        for plugin in plugins:
            self.register_plugin(plugin)

        self.add_event_handler("session_start", self.start)
        self.add_event_handler("message", self.read_message)
        self.add_event_handler("groupchat_subject", self.groupchat_subject)

        self.custom_roster = self.config.get('roster')

        if not self.config.get('lisent'):
            self.config['lisent'] = ['chat', 'normal']
            if self.config.get('rooms', None):
                self.config['lisent'].append('groupchat')

    def start(self, event):
        self.send_presence()
        self.get_roster()
        if self.config.get('vcard', None):
            vcard = self['xep_0054'].stanza.VCardTemp()
            vcard['JABBERID'] = self.boundjid.bare
            for f in self.config['vcard']:
                vcard[f] = self.config['vcard'][f]
                if f.upper() == 'NICKNAME':
                    self.nick = self.config['vcard'][f]
            self['xep_0054'].publish_vcard(vcard)
        if self.config.get('avatar', None):
            avatar_data = None
            try:
                with open(self.config['avatar'], 'rb') as avatar_file:
                    avatar_data = avatar_file.read()
            except IOError:
                logging.debug('Could not load avatar')
            if avatar_data:
                ext = os.path.splitext(self.config['avatar'])[1][1:]
                mtype = 'image/' + ext
                avatar_id = self['xep_0084'].generate_id(avatar_data)
                info = {
                    'id': avatar_id,
                    'type': mtype,
                    'bytes': len(avatar_data)
                }
                self['xep_0084'].publish_avatar(avatar_data)
                self['xep_0084'].publish_avatar_metadata(items=[info])
                self['xep_0153'].set_avatar(avatar=avatar_data, mtype=mtype)

        for room in self.config.get('rooms', []):
            self.plugin['xep_0045'].joinMUC(room,
                                            self.nick,
                                            wait=True)
            msg = self.joined_room(room)
            if msg:
                self.send_message(mto=room, mbody=msg, mtype='groupchat')

        return None

    def joined_room(self, room):
        pass

    def groupchat_subject(self, data):
        pass

    def get_match(self, regex, mode, text):
        if mode == "findall":
            return regex.findall(text)
        m = None
        if mode == "match":
            m = regex.match(text)
        elif mode == "search":
            m = regex.search(text)
        if m:
            return m.groups()
        return None

    def get_cmd(self, msg, user, text):
        delay = self.is_delay(msg)
        cmd = text.split(' ', 1)[0].lower()
        for c in self.commands:
            if delay and not c._command_delay:
                continue
            if c._command_fromuser and user not in c._command_fromuser:
                continue
            if c._command_regex:
                arg = self.get_match(
                    c._command_regex, c._command_rg_mode, text)
                if arg is not None and len(arg) > 0:
                    return c, arg
            elif cmd in c._command_names:
                return c, text.split(' ')[1:]
        return None, None

    def read_message(self, msg):
        if msg['type'] not in self.config['lisent'] or not msg['body'] or not msg['from']:
            return
        if msg['type'] == 'groupchat' and msg['from'].resource.lower() == self.nick.lower():
            return

        user = msg['from'].bare
        if user == self.boundjid.bare or (self.custom_roster and user not in self.custom_roster):
            return

        text = sp.sub(" ", msg['body']).strip()
        if len(text) == 0:
            return

        cmd, args = self.get_cmd(msg, user, text)
        if not cmd:
            self.log.debug("Unknown command from %s: %s" % (user, text))
            return

        self.log.debug("*** Command from %s: %s" % (user, text))

        if msg['type'] == 'groupchat':
            user = msg['from'].resource

        try:
            reply = cmd(*args, user=user, text=text, msg=msg)
        except Exception as error:
            self.log.exception('An error happened while processing '
                               'the message: %s' % text)
            reply = self.command_error(
                error, *args, user=user, text=text, msg=msg)
        if reply:
            self.reply_message(msg, reply, *args, **msg)

    def command_error(self, *args, **kwargs):
        return self.MSG_ERROR_OCCURRED

    def tune_reply(self, txt):
        return txt

    def reply_message(self, msg, txt, *args, **kwargs):
        msgreply = msg.reply(self.tune_reply(txt))
        _to = msgreply['to']
        msgreply.send()
        if self.config.get('img_to_oob', False):
            imgs = set([i[0] for i in url_img.findall(txt)])
            for i in imgs:
                imgreply = msg.reply()
                imgreply['to'] = _to
                imgreply['oob']['url'] = i
                imgreply.send()

    def is_delay(self, msg):
        return self.delay and bool(msg['delay']._get_attr('stamp'))



import re
from os.path import expanduser, isfile

import yaml
from bs4 import BeautifulSoup, Tag
from markdownify import ATX, MarkdownConverter


def get_config(path):
    if path is None:
        raise Exception("Config file is mandatory")
    if isinstance(path, dict):
        config = path
    elif isinstance(path, str):
        path = expanduser(path)
        if not isfile(path):
            raise Exception(path+" doesn't exist")
        with open(path, 'r') as f:
            config = yaml.load(f, Loader=yaml.FullLoader)
    else:
        raise Exception("Config must be str or dict")
    if "user" not in config or "pass" not in config:
        raise Exception("Missing key values")
    return config


re_br = re.compile(r"\s*\n\s*\n\s*")
re_lf = re.compile(r" +$", re.MULTILINE)


class Xep0393Converter(MarkdownConverter):
    """ Create XEP-0393 from html, see https://xmpp.org/extensions/xep-0393.html """

    def __init__(self, *args, **kargv):
        MarkdownConverter.__init__(self, *args, **kargv)

    def convert(self, html):
        if isinstance(html, (Tag, BeautifulSoup)):
            html = html.find("body") or html
            html = str(html)
        for tag in ('u', 'ul', 'ol', "b", "strong", "em", "i", "del", "s", "strike", "span", "a"):
            html = re.sub(r"</" + tag + r">(\s*)<" +
                          tag + "[^>]*>", r"\1", html)
        for tag in ("b", "strong", "em", "i", "del", "s", "strike", "span", "a"):
            for r in (
                re.compile(r"(<"+tag+r"[^>]*>)(\s+)"),
                re.compile(r"(\s+)(</"+tag+r">)")
            ):
                html = r.sub(r"\2\1", html)
        for tag in ('li', 'th', 'td', 'div', 'caption', "p", "div", "h[1-6]"):
            html = re.sub(r"(<"+tag+r"[^>]*>)(\s+)", r"\1", html)
            html = re.sub(r"(\s+)(</"+tag+r">)", r"\2", html)
        text = MarkdownConverter.convert(self, html)
        text = re_br.sub("\n\n", text)
        text = re_lf.sub("", text)
        return text.strip()

    def convert_em(self, el, text):
        return '_%s_' % text if text else ''

    def convert_strong(self, el, text):
        return '*%s*' % text if text else ''

    def convert_s(self, el, text):
        return self.convert_del(el, text)

    def convert_strike(self, el, text):
        return self.convert_del(el, text)

    def convert_del(self, el, text):
        return '~%s~' % text if text else ''


def to_xep0393(html, **options):
    options["heading_style"] = options.get("heading_style", ATX)
    return Xep0393Converter(**options).convert(html)
